
COMMON_MODEL = [
    "source_configs/config.rel",
    "batch_config/batch_config.rel",
    "batch_config/workflow/workflow.rel",
    "batch_config/workflow/steps/configure_sources.rel",
    "batch_config/workflow/steps/export.rel",
    "batch_config/workflow/steps/install_models.rel",
    "batch_config/workflow/steps/invoke_solver.rel",
    "batch_config/workflow/steps/load_data.rel",
    "batch_config/workflow/steps/materialize.rel"
]

COMMON_MODEL_RELATIVE_PATH = "/../rel"

CONFIG_BASE_RELATION = "batch:config"

DATE_FORMAT = "%Y%m%d"

DATE_PREFIX = "data_dt="  # TODO use pattern?

BLOB_PAGE_SIZE = 500

# RAI constants
IMPORT_CONFIG_REL = "import_config"

MISSED_RESOURCES_REL = "missing_resources_json"
WORKFLOW_JSON_REL = "workflow_json"
BATCH_CONFIG_REL = "batch:config"

FILE_LOAD_RELATION = {
    "CSV": "load_csv",
    "JSONL": "load_jsonlines_general",
    "JSON": "load_json" # TODO: maybe load_json_general?
}
# Env config params
RAI_PROFILE = "RAI_PROFILE"
RAI_PROFILE_PATH = "RAI_PROFILE_PATH"
RAI_SDK_HTTP_RETRIES = "RAI_SDK_HTTP_RETRIES"
AZURE_EXPORT_ACCOUNT = "AZURE_EXPORT_ACCOUNT"
AZURE_EXPORT_CONTAINER = "AZURE_EXPORT_CONTAINER"
AZURE_EXPORT_DATA_PATH = "AZURE_EXPORT_DATA_PATH"
AZURE_EXPORT_NUM_FILES = "AZURE_EXPORT_NUM_FILES"
AZURE_EXPORT_SAS = "AZURE_EXPORT_SAS"
AZURE_IMPORT_ACCOUNT = "AZURE_IMPORT_ACCOUNT"
AZURE_IMPORT_CONTAINER = "AZURE_IMPORT_CONTAINER"
AZURE_IMPORT_DATA_PATH = "AZURE_IMPORT_DATA_PATH"
AZURE_IMPORT_SAS = "AZURE_IMPORT_SAS"

# Step parameters
REL_CONFIG_DIR = "rel_config_dir"
OUTPUT_ROOT = "output_root"
LOCAL_DATA_DIR = "local_data_dir"
START_DATE = "start_date"
END_DATE = "end_date"
COLLAPSE_PARTITIONS_ON_LOAD = "collapse_partitions_on_load"
